import os

import law
import luigi
from law.contrib import htcondor

from mldatafind.law.base import DATAFIND_ENV_VARS
from mldatafind.law.parameters import PathParameter


class LDGCondorWorkflow(htcondor.HTCondorWorkflow):
    """
    Base class for law workflows that run via condor on LDG
    """

    condor_directory = PathParameter(
        description="Directory where store condor log files will be written"
    )
    accounting_group_user = luigi.Parameter(
        description="Condors accounting group user name. Defaults to the "
        "`LIGO_USERNAME` environment variable",
        default=os.getenv("LIGO_USERNAME"),
    )
    accounting_group = luigi.Parameter(
        description="Condor accounting group name. Defaults to the "
        "`LIGO_GROUP` environment variable",
        default=os.getenv("LIGO_GROUP"),
    )
    request_disk = luigi.Parameter(
        description="Disk space requested for each job", default="1024"
    )
    request_memory = luigi.Parameter(
        description="Memory requested for each job", default="32678"
    )
    request_cpus = luigi.IntParameter(
        description="Number of cpus requested for each job", default=1
    )

    # don't want to share these between condor tasks via .req()
    exclude_params_req = {"request_memory", "request_disk", "request_cpus"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.htcondor_log_dir.touch()
        self.htcondor_output_directory().touch()

        law.config.update(
            {
                "job": {
                    "job_file_dir_cleanup": "False",
                    "job_file_dir_mkdtemp": "False",
                }
            }
        )

    @property
    def name(self):
        return self.__class__.__name__.lower()

    @property
    def htcondor_log_dir(self):
        return law.LocalDirectoryTarget(self.condor_directory / "logs")

    @property
    def job_file_dir(self):
        return self.htcondor_output_directory().child("jobs", type="d").path

    def htcondor_create_job_file_factory(self, **kwargs):
        # set the job file dir to proper location
        kwargs["dir"] = self.job_file_dir
        return super().htcondor_create_job_file_factory(**kwargs)

    @property
    def law_config(self):
        path = os.getenv("LAW_CONFIG_FILE", "")
        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)
        return path

    def build_environment(self):
        # set necessary env variables for
        # required for LDG data access
        environment = '"'
        for envvar in DATAFIND_ENV_VARS:
            value = os.getenv(envvar)
            if value is not None:
                environment += f"{envvar}={value} "

        # forward current path and law config
        environment += f"PATH={os.getenv('PATH')} "
        environment += f"LAW_CONFIG_FILE={self.law_config} "
        environment += f"USER={os.getenv('USER')} "
        return environment

    def htcondor_output_directory(self):
        return law.LocalDirectoryTarget(self.condor_directory)

    def htcondor_use_local_scheduler(self):
        return True

    def append_memory(self):
        raise NotImplementedError

    def append_logs(self, config):
        for output in ["log", "output", "error"]:
            ext = output[:3]
            config.custom_content.append(
                (
                    output,
                    os.path.join(
                        self.htcondor_log_dir.path,
                        f"{self.name}-$(ProcID).{ext}",
                    ),
                )
            )

    def htcondor_job_config(self, config, job_num, branches):
        environment = self.build_environment()
        config.custom_content.append(("environment", environment))
        config.custom_content.append(("stream_error", "True"))
        config.custom_content.append(("stream_output", "True"))
        config.custom_content.append(
            ("accounting_group", self.accounting_group)
        )
        config.custom_content.append(
            ("accounting_group_user", self.accounting_group_user)
        )
        config.custom_content.append(("request_disk", self.request_disk))
        config.custom_content.append(("request_cpus", self.request_cpus))
        self.append_memory(config)
        self.append_logs(config)
        return config



import law
import luigi
from luigi.util import inherits

from mldatafind.law.base import DataTask
from mldatafind.law.parameters import OptionalPathParameter, PathParameter
from mldatafind.law.targets import s3_or_local
from mldatafind.law.tasks.condor.workflows import StaticMemoryWorkflow
from mldatafind.law.tasks.segments import Query


@inherits(Query)
class Fetch(law.LocalWorkflow, StaticMemoryWorkflow, DataTask):
    """
    Law workflow for fetching strain data
    """

    data_dir = PathParameter(
        description="Directory to store fetched data. "
        "Can be a local path or an s3 path."
    )
    segments_file = OptionalPathParameter(
        description="Path where segments file will be stored. "
        "If not provided, will use the "
        "segments file from the Query task.",
        default="",
    )
    sample_rate = luigi.FloatParameter(
        description="Rate at which fetched data will be sampled in Hz"
    )
    channels = luigi.ListParameter(
        description="List of channels to fetch",
    )
    max_duration = luigi.OptionalFloatParameter(
        description="Maximum duration for each segment in seconds. "
        "Segments of longer length will be split into multiple segments. "
        "Default is no maximum duration.",
        default="",
    )
    resample_method = luigi.Parameter(
        description="Resampling method to use. Either 'gwpy' or 'lal'.",
        default="gwpy",
    )
    prefix = luigi.Parameter(default="background")

    exclude_params_req = {"condor_directory"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not str(self.data_dir).startswith("s3://"):
            self.data_dir.mkdir(exist_ok=True, parents=True)

        if not self.segments_file:
            self.segments_file = self.data_dir / "segments.txt"

    @law.dynamic_workflow_condition
    def workflow_condition(self) -> bool:
        return self.workflow_input()["segments"].exists()

    def load_segments(self):
        with self.workflow_input()["segments"].open("r") as f:
            segments = f.read().splitlines()[1:]
        return segments

    @workflow_condition.create_branch_map
    def create_branch_map(self):
        segments = self.load_segments()
        branch_map, i = {}, 1
        for segment in segments:
            segment = segment.split("\t")
            start, duration = map(float, segment[1::2])
            step = duration if self.max_duration is None else self.max_duration
            num_steps = (duration - 1) // step + 1

            for j in range(int(num_steps)):
                segstart = start + j * step
                segdur = min(start + duration - segstart, step)
                branch_map[i] = (segstart, segdur)
                i += 1
        return branch_map

    def workflow_requires(self):
        reqs = super().workflow_requires()
        reqs["segments"] = Query.req(self, segments_file=self.segments_file)
        return reqs

    @workflow_condition.output
    def output(self):
        start, duration = self.branch_data
        start = int(float(start))
        duration = int(float(duration))
        fname = "{}-{}-{}.hdf5".format(self.prefix, start, duration)
        fname = self.data_dir / fname
        return s3_or_local(fname)

    def run(self):
        import h5py

        from mldatafind.fetch import fetch

        start, duration = self.branch_data
        start = int(float(start))
        duration = int(float(duration))

        X = fetch(
            start,
            start + duration,
            self.channels,
            self.sample_rate,
            resample_method=self.resample_method,
        )
        size = int(duration * self.sample_rate)
        with self.output().open("w") as f:
            with h5py.File(f, "w") as h5file:
                # write with chunking for dataloading perf increase
                X.write(
                    h5file,
                    format="hdf5",
                    chunks=(min(size, 131072),),
                    compression=None,
                )



from typing import List, Literal
import numpy as np
import lal
from gwpy.timeseries import TimeSeries, TimeSeriesDict

from mldatafind.authenticate import authenticate

# channel names that signal to fetch open data
OPEN_DATA_CHANNELS = ["H1", "L1", "V1"]


def _fetch_open_data(
    ifos: List[str], start: float, end: float, **kwargs
) -> TimeSeriesDict:
    ts_dict = TimeSeriesDict()
    for ifo in ifos:
        ts_dict[ifo] = TimeSeries.fetch_open_data(ifo, start, end, **kwargs)
    return ts_dict


def fetch(
    start: float,
    end: float,
    channels: list[str],
    sample_rate: float,
    nproc: int = 3,
    verbose: bool = True,
    allow_tape: bool = True,
    resample_method: Literal["gwpy", "lal"] = "gwpy",
) -> TimeSeriesDict:
    """
    Simple wrapper to annotate and simplify
    the kwargs so that jsonargparse can build
    a parser out of them.
    """

    open_data_channels = list(
        filter(lambda x: x in OPEN_DATA_CHANNELS, channels)
    )
    channels = list(filter(lambda x: x not in OPEN_DATA_CHANNELS, channels))

    # fetch data from nds2
    data = TimeSeriesDict()
    if channels:
        # only auth if we have non open channels to fetch
        authenticate()
        for channel in channels:
            ifo = channel.split(":")[0]
            data[ifo] = TimeSeries.get(
                channel,
                start=start,
                end=end,
                verbose=verbose,
                nproc=nproc,
                allow_tape=allow_tape,
            )

    # fetch open data channels and combine
    if open_data_channels:
        open_data_ts_dict = _fetch_open_data(
            open_data_channels,
            start=start,
            end=end,
            verbose=verbose,
            nproc=nproc,
        )
        data.update(open_data_ts_dict)

    if resample_method == "gwpy":
        data = data.resample(sample_rate)
    elif resample_method == "lal":
        for key, ts in data.items():
            lal_data = ts.astype(np.float64).to_lal()
            lal.ResampleREAL8TimeSeries(lal_data, float(1 / sample_rate))
            lal_data = TimeSeries(
                lal_data.data.data,
                epoch=lal_data.epoch,
                dt=lal_data.deltaT,
            )
            data[key] = lal_data
    else:
        raise ValueError(f"Invalid resampling method {resample_method}")

    return data



import luigi

from mldatafind.law.base import DataTask
from mldatafind.law.parameters import PathParameter
from mldatafind.law.targets import s3_or_local


class Query(DataTask):
    """
    Law task to query data quality segments
    """

    start = luigi.FloatParameter(description="Start time of segments to query")
    end = luigi.FloatParameter(description="End time of segments to query")
    segments_file = PathParameter(
        description="Output path where segments are written",
    )
    ifos = luigi.ListParameter(
        description="List of ifos for which "
        "to query segments, and strain data. "
    )
    flags = luigi.ListParameter(
        description="Data quality flag to query for each ifo. "
        "If '{ifo}_DATA', will query for open data segments. "
        "Expect one value per ifo in `ifos`. "
    )
    min_duration = luigi.OptionalFloatParameter(
        description="Minimum duration of segments to query. "
        "Any segments of shorter length will be discarded",
        default="",
    )

    retry_count = 5

    def output(self):
        return s3_or_local(self.segments_file, format="txt")

    def run(self):
        from mldatafind.segments import DataQualityDict

        segments = DataQualityDict.query_segments(
            self.flags,
            self.start,
            self.end,
            self.min_duration,
        )
        with self.output().open("w") as f:
            segments.write(f, format="segwizard")



import os
import shutil
import subprocess

_kinit_errs = {
    "Key table file '{keytab_location}' not found": (
        "kinit command failed because key table file "
        "'{keytab_location}' not found. See instructions for "
        "setting up passwordless authentication at "
        "https://computing.docs.ligo.org/guide/auth/kerberos/#ligo"
    ),
    "Keytab contains no suitable keys for {user}@LIGO.ORG": (
        "kinit command failed because key table file "
        "was not configured for user {user}. Make sure "
        "that when following instructions at "
        "https://computing.docs.ligo.org/guide/auth/kerberos/#ligo "
        "that you replace 'albert.einstein' with '{user}'."
    ),
    "Password incorrect while getting initial credentials": (
        "kinit command failed because password saved to "
        "key table file is incorrect for user {user}. Re-generate "
        "with the correct password using the instructions at "
        "https://computing.docs.ligo.org/guide/auth/kerberos/#ligo"
    ),
}


def _validate_env(env_var: str):
    """
    Validate the existence of an environment variable
    Args:
        env_var : str
            The name of the environment variable
    Raises:
        ValueError If the environment variable is not set

    Returns The value of the environment variable
    """
    value = os.getenv(env_var)
    if value is None:
        raise ValueError(f"{env_var} environment variable not set")

    return value


def _check_kinit_errs(stderr: str, user: str, keytab_location: str) -> None:
    for key, msg in _kinit_errs.items():
        key = key.format(user=user, keytab_location=keytab_location)
        if stderr.startswith("kinit: " + key):
            msg = msg.format(user=user, keytab_location=keytab_location)
            raise OSError(msg)


def kinit():
    user = _validate_env("LIGO_USERNAME")
    keytab_location = _validate_env("KRB5_KTNAME")

    kinit_command = shutil.which("kinit")
    if kinit_command is None:
        raise ValueError("kinit command not found")

    args = [
        kinit_command,
        "-p",
        f"{user}@LIGO.ORG",
        "-k",
        "-t",
        keytab_location,
    ]

    response = subprocess.run(args, capture_output=True, text=True)
    if response.returncode:
        # first check to see if we recognize any of the issues
        _check_kinit_errs(response.stderr, user, keytab_location)

        # if not, raise a generic error with kinit's full stderr
        raise RuntimeError(
            "kinit command failed with return code {}: {}".format(
                response.returncode, response.stderr
            )
        )


def make_cert(cert_path: str) -> None:
    from ciecplib.ui import get_cert
    from ciecplib.x509 import write_cert

    kinit()
    cert, key = get_cert(kerberos=True)
    write_cert(cert_path, cert, key)


def authenticate():
    """
    Authenticate a user to access LIGO data sources

    This function will load a X509 certificate from the environment
    variable `X509_USER_PROXY`. If the credential doesn't exist,
    it will create a new one. If the credential exists and is valid,
    it will continue to use it. Otherwise, it will generate a new credential.

    If generating new credential, a kerberos keytab is required
    for passwordless authentication. It's location should be
    specified in the environment variable `KRB5_KTNAME`.
    This function assumes the user has already generated a kerberos keytab
    with principal user.name@LIGO.ORG. This function will read in username
    from the environment variable `LIGO_USERNAME`

    For instructions on generating a kerberos keytab,
    see https://computing.docs.ligo.org/guide/auth/kerberos/

    """

    from ciecplib.x509 import check_cert, load_cert

    cert_path = _validate_env("X509_USER_PROXY")
    if os.path.exists(cert_path):
        cert = load_cert(cert_path)
        try:
            check_cert(cert)
        except RuntimeError:
            make_cert(cert_path)
    else:
        make_cert(cert_path)

    return cert_path



from typing import Iterable, Optional

from gwpy.segments import DataQualityDict, DataQualityFlag, SegmentList

from mldatafind.authenticate import authenticate

OPEN_DATA_FLAGS = ["H1_DATA", "L1_DATA", "V1_DATA"]
O3A_END = 1253977218
O3B_START = 1256655618


class DataQualityDict(DataQualityDict):
    @classmethod
    def query_non_open(
        cls, flags: Iterable[str], start: float, end: float, **kwargs
    ) -> DataQualityDict:
        try:
            return cls.query_dqsegdb(flags, start, end, **kwargs)
        except OSError as e:
            if not str(e).startswith(
                "Could not find the TLS certificate file"
            ):
                # TODO: what's the error for an expired certificate?
                raise

            return cls.query_dqsegdb(flags, start, end, **kwargs)

    @classmethod
    def query_open(
        cls, flags: Iterable[str], start: float, end: float, **kwargs
    ) -> DataQualityDict:
        dqdict = cls()
        for flag in flags:
            dqdict[flag] = DataQualityFlag.fetch_open_data(
                flag, start, end, **kwargs
            )
        return dqdict

    @classmethod
    def _query_segments(
        cls,
        flags: Iterable[str],
        start: float,
        end: float,
        min_duration: Optional[float] = None,
        **kwargs,
    ) -> SegmentList:
        flags = set(flags)
        open_flags = set(OPEN_DATA_FLAGS)

        open_data_flags = list(flags & open_flags)
        flags = list(flags - open_flags)

        segments = cls()
        if flags:
            # only auth if we have non open flags to fetch
            authenticate()
            segments.update(cls.query_non_open(flags, start, end, **kwargs))
        if open_data_flags:
            segments.update(
                cls.query_open(open_data_flags, start, end, **kwargs)
            )

        segments = segments.intersection().active
        if min_duration is not None:
            segments = filter(lambda i: i[1] - i[0] >= min_duration, segments)
            segments = SegmentList(segments)
        return segments

    @classmethod
    def query_segments(
        cls,
        flags: Iterable[str],
        start: float,
        end: float,
        min_duration: Optional[float] = None,
        **kwargs,
    ) -> SegmentList:
        # if the requested time period
        # spans O3a to O3b, query the two
        # separately and append
        if start < O3A_END and end > O3B_START:
            segments = SegmentList()
            segments.extend(
                cls._query_segments(
                    flags, start, O3A_END, min_duration, **kwargs
                )
            )
            segments.extend(
                cls._query_segments(
                    flags, O3B_START, end, min_duration, **kwargs
                )
            )
            return segments
        # otherwise, just query the whole period
        return cls._query_segments(flags, start, end, min_duration, **kwargs)



from pathlib import Path
from typing import Union, get_args

import luigi
from cloudpathlib import CloudPath
from luigi.parameter import OptionalParameterMixin

PATH_LIKE = Union[CloudPath, Path, str]


class PathParameter(luigi.Parameter):
    """
    luigi `Parameter` class that handles parsing strings
    into pathlib.Path or cloudpathlib.S3Path objects.
    """

    def parse(self, x: PATH_LIKE):
        if isinstance(x, (Path, CloudPath)):
            return x / ""
        if isinstance(x, str):
            if x.startswith("s3://"):
                return CloudPath(x) / ""
            else:
                return Path(x) / ""
        else:
            raise ValueError(
                f"Expected string, Path, or CloudPath, got {type(x)}"
            )

    def serialize(self, x):
        return str(x)

    def normalize(self, x):
        return self.parse(x)


class OptionalPathParameter(OptionalParameterMixin, PathParameter):
    expected_type = get_args(PATH_LIKE)
    pass



from typing import Literal

import luigi
from luigi.contrib.s3 import S3Target
from luigi.format import BaseWrapper, WrappedFormat

from mldatafind.law.config import s3
from mldatafind.law.parameters import PATH_LIKE


# format for writing h5 files
# to s3 via bytes streams
class BytesFormat(WrappedFormat):
    input = "bytes"
    output = "bytes"
    wrapper_cls = BaseWrapper


Bytes = BytesFormat()


# law targets compatible with luigi S3 targets
class LawS3Target(S3Target):
    optional = False

    def complete(self):
        return self.exists()


class LawLocalTarget(luigi.LocalTarget):
    optional = False

    def complete(self):
        return self.exists()


def s3_or_local(path: PATH_LIKE, format: Literal["hdf5", "txt"] = "hdf5"):
    """
    Dynamically return a LawS3Target or LawLocalTarget depending on
    the specified path.

    Args:
        path:
            The path to the target. If the path starts with "s3://",
            a LawS3Target is returned, otherwise a LawLocalTarget is returned.
        format:
            The format of the target. Can be either 'hdf5' or 'txt'.
            Defaults to "hdf5".

    Returns:
        A LawS3Target or LawLocalTarget
    """
    format = Bytes if format == "hdf5" else None
    path = str(path)
    if path.startswith("s3://"):
        return LawS3Target(
            path,
            client=s3().client,
            ContentType="application/octet-stream",
            format=format,
        )
    else:
        return LawLocalTarget(path, format=format)



import luigi

from mldatafind.law.tasks.condor.base import LDGCondorWorkflow


class StaticMemoryWorkflow(LDGCondorWorkflow):
    """
    Workflow that requests a fixed amount of memory for each job
    """

    def append_memory(self, config):
        config.custom_content.append(("request_memory", self.request_memory))


class DynamicMemoryWorklow(LDGCondorWorkflow):
    """
    Workflow that dynamically updates memory
    based on the memory usage of the job
    """

    max_memory = luigi.Parameter(default="7G")

    def append_memory(self, config):
        config.custom_content.append(
            ("+InitialRequestMemory", self.request_memory)
        )
        config.custom_content.append(
            (
                "request_memory",
                f"ifthenelse(isUndefined(MemoryUsage), {self.request_memory}, int(3*MemoryUsage))",  # noqa
            )
        )
        config.custom_content.append(
            (
                "periodic_release",
                "(HoldReasonCode =?= 26 || HoldReasonCode =?= 34) && (JobStatus == 5)",  # noqa
            )
        )
        config.custom_content.append(
            (
                "periodic_remove",
                f"(JobStatus == 1) && MemoryUsage >= {self.max_memory}",
            )
        )



import os
from pathlib import Path

import law
import luigi
from law.contrib import singularity
from law.contrib.singularity.config import config_defaults

root = Path(__file__).resolve().parent.parent.parent

DATAFIND_ENV_VARS = [
    "KRB5_KTNAME",
    "X509_USER_PROXY",
    "GWDATAFIND_SERVER",
    "NDSSERVER",
    "LIGO_USERNAME",
    "DEFAULT_SEGMENT_SERVER",
    "AWS_ENDPOINT_URL",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_ACCESS_KEY_ID",
]


class DataSandbox(singularity.SingularitySandbox):
    """
    Singularity sandbox for running mldatafind tasks
    """

    sandbox_type = "mldatafind"

    def get_custom_config_section_postfix(self):
        return self.sandbox_type

    @classmethod
    def config(cls):
        config = {}
        default = config_defaults(None).pop("singularity_sandbox")
        default["law_executable"] = "/opt/env/bin/law"
        default["forward_law"] = False
        postfix = cls.sandbox_type
        config[f"singularity_sandbox_{postfix}"] = default
        return config

    @property
    def data_directories(self):
        """
        Data directories on LDG clusters to bind to the container
        to enable local data discovery
        """
        return map(
            Path, ["/cvmfs", "/hdfs", "/gpfs", "/ceph", "/hadoop", "/archive"]
        )

    def _get_volumes(self):
        volumes = super()._get_volumes()
        # bind data directories if they
        # exist on the local cluster
        for dir in self.data_directories:
            if dir.exists():
                volumes[str(dir)] = str(dir)

        # bind users /local directory for
        # storing large tmp files,
        # e.g. for local storage before
        # being dumped to s3 by luigi
        local = Path(f"/local/{os.getenv('USER')}")
        if local.exists():
            local = str(local)
            volumes[local] = local

        # bind aws directory that contains s3 credentials
        aws_dir = os.path.expanduser("~/.aws/")
        volumes[aws_dir] = aws_dir
        return volumes

    def _get_env(self):
        env = super()._get_env()
        for envvar in DATAFIND_ENV_VARS:
            value = os.getenv(envvar)
            if value is not None:
                env[envvar] = value
        return env


law.config.update(DataSandbox.config())


class DataTask(law.SandboxTask):
    """
    law SandboxTask for running mldatafind workflows
    """

    image = luigi.PathParameter(
        default=os.getenv("MLDATAFIND_CONTAINER", ""),
        significant=False,
        description="Path to the singularity container to use for the task. "
        "Defaults to the `MLDATAFIND_CONTAINER` environment variable. ",
    )



import os

import luigi
from luigi.contrib.s3 import S3Client


class s3(luigi.Config):
    """
    Global S3 client configuration.
    """

    endpoint_url = luigi.Parameter(
        description="S3 endpoint URL. Defaults to reading from "
        "`AWS_ENDPOINT_URL` environment variable.",
        default=os.getenv("AWS_ENDPOINT_URL"),
    )
    aws_access_key_id = luigi.Parameter(
        description="AWS access key ID. Defaults to reading from "
        "`AWS_ACCESS_KEY_ID` environment variable.",
        default=os.getenv("AWS_ACCESS_KEY_ID"),
    )
    aws_secret_access_key = luigi.Parameter(
        description="AWS secret access key. Defaults to reading from "
        "`AWS_SECRET_ACCESS_KEY` environment variable.",
        default=os.getenv("AWS_SECRET_ACCESS_KEY"),
    )

    @property
    def client(self):
        return S3Client(endpoint_url=self.endpoint_url)
